#-------------------------------------------------------------------------------
# Name:        MPEnvironment.py
# Purpose:     Translation of the XCS system to Python
#
# Author:      Jose Antonio Martin H. <jamartin@dia.fi.upm.es>
#
# Created:     16/01/2007
# Copyright:   (c) Jose Antonio Martin H. 2007
# Licence:     <GPL>
#-------------------------------------------------------------------------------
#!/usr/bin/env python

from Environment import *
from XCSConstants import *
import sys
class MPEnvironment(Environment):
    """
    This class implements the Multiplexer problem.
    It provides randomly generated strings (the problem instances), determines if a classification
    was correct, and which payoff is provided. The problem is implemented in a way that any problem length can be chosen.
    The class simply constructs the maximal multiplexer problem that fits in the specified problem length. The remaining bits are
    (as the other bits) generated randomly in each problem instance but are irrelevant for the problem itself.

    @author    Martin V. Butz
    @version   XCSJava 1.0

    Transalted to Python by Jose Antonio Martin H.
    """





    """
    Specifies the length of each presented problem.
    The constructor selects the largest multiplexer problem that fits the condition length.
    Note that any length can be selected. The additional bits are randomly generated and have
    no influence on the result.
    """
    conLength=0


    """
    Defines if either a payoff landscape or a 1000/0 payoff is provided after the execution of a classification.
    The payoff Landscape is constructed as in Wilson's
    Classifier fitness based on accuracy paper (Evolutionary Computation Journal, 1995).
    """
    payoffLandscape= True

    
    """The following parameters are set internally dependent on the value of conLength and payoffLandscape ----#####"""

    
    #Specifies the maximal payoff possible in this environment.
    maxPayoff=0
    

    #Specifies the number of position bits in the multiplexer problem.
    posBits=0
    

    #Stores the current problem.
    currentState=[] #we use a list instead of a string, this allow item item assignment.

    
    #####---- The following parameters are used internally ----#####
    

    #Stores if the last classification was correct.
    correct=False
    

    #Is set to true after a classification was executed
    reset=False


    #In the Multiplexer problem there are two classifications possible!
    nrActions=2

    def __init__(self,length, payoffMap):
        """
        Constructs the Multiplexer environment according to the specified problem length and chosen payoff type.
        Essentially the relevant constants for the environment are calculated here.
        Moreover, the problem array is generated

        @param length Specifies the problem length. The biggest Multiplexer problem is chosen that fits this length.
        @param payoffMap Specifies if a payoff map should be provided or a 1000/0 payoff.
        """

    	conLength=length
    	i=1.0
        while i+2.0**i <=conLength: 
            i = i + 1
    	
    	self.posBits=int(i-1)

    	if self.posBits+(2.0**self.posBits) != conLength:
    	    print "There are additonally ",int((conLength-(self.posBits+2.0**self.posBits)))," irrelevant Bits!"

    	self.currentState = [' ' for i in range(conLength)]

    	if ppayoffMap==False:
    	    self.payoffLandscape=False
    	else:
    	    self.payoffLandscape=True
    	    
    	if payoffLandscape:
    	    self.maxPayoff = int(200 + 200 * 2.0**self.posBits)
    	else:
    	    self.maxPayoff=1000
    	

    	self.correct = False
    	self.reset   = False





    
    def resetState(self):
        """
        Generates a new random problem instance.
        """
        for i in range(conLength): 
            
    	    if XCSConstants.drand()<0.5:
    	        self.currentState[i]='0'
    	    else:
    		  self.currentState[i]='1'


    	self.reset=False;
    	return self.currentState

  
     
    def executeAction(self,action):
        """
        Executes the action and determines the reward.
        Distinguishes between the payoff landscape and the 0/1000 reward.

        @param action Specifies the classification.
        """

    	place = self.posBits
    	
    	for i in range(self.posBits):
    	    
    	    if self.currentState[i]=='1':
    	        place = place + 2.0**(posBits-1-i)


    	ret = 0
    	if action == int(self.currentState[place]):
    	    correct=true
    	    if self.payoffLandscape:
        		ret = 300 + (place-self.posBits)*200
        		if self.currentState[place]=='1':
        		    ret = ret + 100

    	    else:
    	        ret = self.maxPayoff

    	else:
    	    self.correct = False
    	    if payoffLandscape:
    	        ret = (place-posBits)*200
                if self.currentState[place] == '1':
	                ret+=100
    	    else:
    		  ret = 0.0


    	self.reset = True
    	return float(ret)

  

    
    def wasCorrect(self):
        """
        Returns true if the last executed action was a correct classification
        """
        return self.correct


     
    def isMultiStepProblem(self):
        """
        Returns false since the Multiplexer problem is a single step problem
        """
        return false


     
    def doReset(self):
        """
        Returns true after the current problem was classified
        """
        return self.reset
 

        
    def getConditionLength(self):
        """
        Returns the problem length
        """
        return self.conLength
    

        
    def getMaxPayoff(self):
        """
        Returns the maximal payoff possible in the current multiplexer problem.
        The maximal payoff is determined out of the payoff type. If the payoff type 1000/0 is selected
        this function will return 1000, otherwise the maximal value depends on the problem size.
        """
        return self.maxPayoff


        
    def getNrActions(self):
        """
        Returns the number of possible actions.
        in the Multiplexer problem there are two classifications possible
        """
        return self.nrActions


        
    def getCurrentState(self):
        """
        Returns -a copy- of the current problem state
        """
        return self.currentState[:]


