#-------------------------------------------------------------------------------
# Name:        MazeEnvironment.py
# Purpose:     Translation of the XCS system to Python
#
# Author:      Jose Antonio Martin H. <jamartin@dia.fi.upm.es>
#
# Created:     16/01/2007
# Copyright:   (c) Jose Antonio Martin H. 2007
# Licence:     <GPL>
#-------------------------------------------------------------------------------
#!/usr/bin/env python
from Environment import *
from XCSConstants import *
import math
import sys

class MazeEnvironment(Environment):
    """
    This class implements a maze environment.
    It reads in a maze from a file. This file must contain the maze in a nxm matrix where each
    entry specifies one distinct position in the maze. If the animat reaches an edge and moves out of the edge,
    it will reenter the maze from the other side (if not blocked). Obstacles are coded as 'O'or'0' and 'Q', empty positions as
    '.'or'*', and food as 'F' and 'G'. A two bit or three bit coding can be chosen for the coding of the perceptions. In the
    three bit coding the perceptions differentiate 'F' and 'G', and 'O'/'0' and 'Q' while they appear the same in the two
    bit coding.
    Eight movements are possible to the adjacent cells (if not blocked). The environment starts by positioning the animat
    at a random position. It moves as long as no food position is reached. If a food position is reached, the environment
    provides the payoff specified in maxPayoff. Also, the reset flag is set to true in this case.

    @author    Martin V. Butz
    @version   XCSJava 1.0

    Transalted to Python by Jose Antonio Martin H.
    """
    
    
    

    """###################---- Constants which are preset ----###################"""
    

    #The number of perceptions of the animat.

    conLength = 8


    #The number of generally possible movements.
    nrActions = 8


    #The payoff provided at a food position.
    maxPayoff = 1000
    

    #The binary code of an empty position (000).
    freeAtt = ['0','0','0']
    

    #The binary code of a food-F position (110).
    foodF = ['1','1','0']
    

    #The binary code of a food-G position (111).
    foodG = ['1','1','1']
    

    #The binary code of an O-obstacle position (010).
    obstacleO = ['0','1','0']
    

    #The binary code of a Q-obstacle position (011).
    obstacleQ = ['0','1','1']



    """###################---- These variables are set in the constructor ----###################"""

    """
    The attribute Length specifies the number of bits that code each perceived position in the maze.
    It can be set to 2 or 3.
    """
    attributeLength = 0


    #The matrix that codes the maze (in binary)
    maze = []


    #The size of the maze in positions.
    xsize = 0
    ysize = 0


    #The current position of the animat.
    xcurrent = 0
    ycurrent = 0


    #Flag which is set to true when food was reached.
    reset = False

        
    def __init__(self, inFileString, attLength):
        """
        The constructor reads in the specified maze file and sets its global parameters accordingly.

        @param inFileString must specify the file name where the maze is coded
        @param attLength specifies the number of bits that specify one attribute (either two or three).
               if wrongly specified, then it is set to two.
        """

    	if attLength==3:
    	    self.attributeLength=3
    	else:
    	    self.attributeLength=2

    	
    	mazeLines=[]
    	try:
    	    fr = open(inFileString)
    	    self.xsize=0
    	    self.ysize=0
    	    for ExtLine in fr:
    	        
    	        Line = ExtLine.strip()
                print Line
                if self.xsize==0:
        		    self.xsize = len(Line)
        		    
                oneLine= ['' for i in range(self.xsize*self.attributeLength)]
    	        for i in range(self.xsize):

                    att = Line[i]
                    
                    if att=='.' or att=='*':
                        for j in range(self.attributeLength):
                            oneLine[i*self.attributeLength + j]=self.freeAtt[j]

                    elif att=='F':
                        for j in range(self.attributeLength):
                            oneLine[i*self.attributeLength + j]=self.foodF[j]

                    elif att=='G':
                        for j in range(self.attributeLength):
                            oneLine[i*self.attributeLength + j]=self.foodG[j]

                    elif att== 'O' or att=='0':
                        for j in range(self.attributeLength):
                            oneLine[i*self.attributeLength + j]=self.obstacleO[j]

                    elif att=='Q':
                        for j in range(self.attributeLength):
                            oneLine[i*self.attributeLength + j]=self.obstacleQ[j]

                    else:
                        raise NameError('Unknown Character:'+att)

                mazeLines.append(oneLine)
                
        		
        except IOError, (errno, strerror):
            print "Could not Read File!"
            print "I/O error(%s): %s" % (errno, strerror)
            raise
        except ValueError:
            print "Could not convert data to an integer."
            raise
        except:
            print "Unexpected error:", sys.exc_info()[0]
            raise

    	
    	self.ysize = len(mazeLines)
    	self.maze= [['' for j in range(self.ysize)] for i in range(self.xsize*self.attributeLength)]

        print "Maze size: ",self.xsize,"x",self.ysize
        for i in range(self.ysize):
    	    line = mazeLines[i]
    	    for j in range(len(line)): 
    	        self.maze[j][i]=line[j]
        
    	self.reset=False
    	self.setRandomPosition()
    	


        
    def setRandomPosition(self):
        """
        Sets the animat to a randomly selected empty position.
        """
    	self.xcurrent=int(XCSConstants.drand()*self.xsize)
    	self.ycurrent=int(XCSConstants.drand()*self.ysize)
    	
        
    	while self.maze[self.xcurrent*self.attributeLength][self.ycurrent]!=self.freeAtt[0] or self.maze[self.xcurrent*self.attributeLength+1][self.ycurrent]!=self.freeAtt[1]:
    	    self.xcurrent=int(XCSConstants.drand()*self.xsize)
    	    self.ycurrent=int(XCSConstants.drand()*self.ysize)
                  
    	           


        
    def getPerceptions(self):
        """
        Returns a String of the perceptions in the current position.
        """
        atl = self.attributeLength
        
    	perc= ['' for i in range(self.getConditionLength())]  # char[getConditionLength()];
    	for i in range(self.conLength): #(int i=0; i<conLength; i++){
    	    for j in range(self.attributeLength): #(int j=0; j<attributeLength; j++){
    	        perc[i*atl + j] = self.maze[j + atl * ((( self.xcurrent + int(((math.ceil(float(i-3.0)/4.0)*2.0)-1.0)*(-1.0)* math.ceil(float(i%4)/4.0)))+self.xsize)%self.xsize)][((self.ycurrent + (int)(((math.ceil(math.floor(float(i%7)/2.0)/4.0)*2.0)-1.0)*math.ceil(float((i+2)%4)/4.0)))+self.ysize)%self.ysize]


    	return perc


        
    def resetState(self):
        """
        Resets the animat to a random empty position and returns the perceptions in this position.
        """
    	self.setRandomPosition()
    	self.reset=False
    	return self.getPerceptions()

  
        
    def getCurrentState(self):
        """
        Returns the current perceptions.
        """
    	return self.getPerceptions()


        
    def executeAction(self, action):
        """
        Executes the specified action in the environment and returns possible payoff.

        @param action The action to be executed.
        """
    	if action<0 or action>7:
    	    print "Not an action!"
    	    raise NameError("Not an action! "+str(action))

    	#Test if new position is empty!!!
    	xaim=((self.xcurrent + int(((math.ceil(float(action-3)/4.0)*2)-1)*(-1)* math.ceil(float(action%4)/4.0)))+self.xsize)%self.xsize
    	yaim=((self.ycurrent + int(((math.ceil(math.floor(float(action%7)/2.0)/4.0)*2)-1.0)*math.ceil(float((action+2)%4)/4.0)))+self.ysize) % self.ysize

        if self.maze[xaim*self.attributeLength][yaim]!=self.obstacleO[0] or self.maze[1+ xaim*self.attributeLength][yaim]!=self.obstacleO[1]:
            self.xcurrent=xaim
    	    self.ycurrent=yaim
    	    if self.maze[self.xcurrent*self.attributeLength][self.ycurrent] == self.foodF[0] and self.maze[1+ self.xcurrent*self.attributeLength][self.ycurrent]==self.foodF[1]:
    		  self.reset=True
    		  return self.maxPayoff

    	    return 0.0

    	return 0.0


        
    def wasCorrect(self):
        """
        Returns always false since there is no real correct or wrong action in the maze environment.
        """

    	return False #No correct or wrong action in this environment


        
    def isMultiStepProblem(self):
        """
        Returns true since any maze is a multi-step environment.
        """
    	return True


        
    def doReset(self):
        """
        Returns if the animat should be reseted.
        The reset flag is set to true once the animat reached a food position.
        """

    	return self.reset


        
    def getConditionLength(self):
        """
        Returns the length of the perceptions.
        """
    	return self.conLength * self.attributeLength


        
    def getMaxPayoff(self):
        """
        Returns the maximal Payoff in the maze.
        """
        return self.maxPayoff


        
    def getNrActions(self):
        """
        Returns the number of actions possible in the maze environment.
        """
        return self.nrActions

