#-------------------------------------------------------------------------------
# Name:        PredictionArray.py
# Purpose:     Translation of the XCS system to Python
#
# Author:      Jose Antonio Martin H. <jamartin@dia.fi.upm.es>
#
# Created:     16/01/2007
# Copyright:   (c) Jose Antonio Martin H. 2007
# Licence:     <GPL>
#-------------------------------------------------------------------------------
#!/usr/bin/env python
from XCSConstants import *
from XClassifierSet import *

import sys
class PredictionArray:
    """
    This class generates a prediction array of the provided set.
    The prediction array is generated according to Wilson's Classifier Fitness Based on Accuracy
    (Evolutionary Computation Journal, 1995).
    Moreover, this class provides all methods to handle selection in the prediction array, essentially, to select the
    best action, a present random action or an action by roulette wheel selection.

    @author    Martin V. Butz
    @version   XCSJava 1.0

    Transalted to Python by Jose Antonio Martin H.
    """

    
    #The prediction array.
    pa = []


    #The sum of the fitnesses of classifiers that represent each entry in the prediction array.
    nr = []
    
    
    def __init__(self,set, size):
        """
        Constructs the prediction array according to the current set and the possible number of actions.

        @param set The classifier set out of which a prediction array is formed (normally the match set).
        @param size The number of entries in the prediction array (should be set to the number of possible actions in the problem)
        """
        
        

    	self.pa = [0.0 for i in range(size)]
    	self.nr = [0.0 for i in range(size)]

        for cl in set.clSet:
    	    self.pa[cl.getAction()] = self.pa[cl.getAction()] + (cl.getPrediction()*cl.getFitness())
    	    self.nr[cl.getAction()] = self.nr[cl.getAction()] + cl.getFitness()

    	for i in range(size):
    	    if self.nr[i]!=0:
    		  self.pa[i] = self.pa[i]/self.nr[i]
    	    else:
    		  self.pa[i]=0.0

        
    def getBestValue(self):
        """
        Returns the highest value in the prediction array.
        """
        return max(self.pa)
    
        
    def getValue(self, i):
        """
        Returns the value of the specified entry in the prediction array.
        """
    	if i>=0 and i<len(self.pa):
    	    return self.pa[i]
    	return -1.0

    ##*************** Action selection functions ****************

        
    def randomActionWinner(self):
        """
        Selects an action randomly.
        The function assures that the chosen action is represented by at least one classifier.
        """
        
        
    	ret=0
    	while True:
    	    ret = int(XCSConstants.drand()*len(self.pa)) 
    	    if self.nr[ret]!=0:
    	        break
    	    
    	return ret
    
    

    def bestActionWinner(self):
        """
        Selects the action in the prediction array with the best value.
        """
              #simply argmax(self.pa) when using numpy
        return self.pa.index(max(self.pa))
    
    
        
    def rouletteActionWinner(self):
        """
        Selects an action in the prediction array by roulette wheel selection.
        """
    	bidSum = sum(self.pa) * XCSConstants.drand()
    	bidC=0.0
    	i = 0
    	while bidC<bidSum:
    	    bidC = bidC + self.pa[i]
            i = i + 1

        return i

