#-------------------------------------------------------------------------------
# Name:        XCSConstants.py
# Purpose:     Translation of the XCS system to Python
#
# Author:      Jose Antonio Martin H. <jamartin@dia.fi.upm.es>
#
# Created:     16/01/2007
# Copyright:   (c) Jose Antonio Martin H. 2007
# Licence:     <GPL>
#-------------------------------------------------------------------------------
#!/usr/bin/env python

class CXCSConstants:

    """
     * This class provides all relevant learning parameters for the XCS as well as
     * other experimental settings and flags. Most parameter-names are chosen similar to
     * the 'An Algorithmic Description of XCS' ( Butz&Wilson, IlliGAL report 2000017).
     *
     * @author    Martin V. Butz
     * @version   XCSJava 1.0
     *

       Transalted to Python by Jose Antonio Martin H.
    """
    
    
    """
      Specifies the maximal number of micro-classifiers in the population.
      In the multiplexer problem this value is set to 400, 800, 2000 in the 6, 11, 20 multiplexer resp..
      In the Woods1 and Woods2 environment the parameter was set to 800.
     """
    maxPopSize=800

    """*
     * The fall of rate in the fitness evaluation.
    """
    alpha=0.1

    """*
     * The learning rate for updating fitness, prediction, prediction error, 
     * and action set size estimate in XCS's classifiers.
     """
    beta=0.2

    """*
     * The discount rate in multi-step problems.
     """
    gamma=0.95

    """*
     * The fraction of the mean fitness of the population below which the fitness of a classifier may be considered 
     * in its vote for deletion.
     """
    delta=0.1

    """*
     * Specifies the exponent in the power function for the fitness evaluation.
     """
    nu=5.
    
    """*
     * The threshold for the GA application in an action set.
     """
    theta_GA=25

    """*
     * The error threshold under which the accuracy of a classifier is set to one.
     """
    epsilon_0=10

    """*
     * Specified the threshold over which the fitness of a classifier may be considered in its deletion probability.
     """
    theta_del=20

    """*
     * The probability of applying crossover in an offspring classifier.
     """
    pX=0.8

    """*
     * The probability of mutating one allele and the action in an offspring classifier.
     """
    pM=0.04
  
    """*
     * The probability of using a don't care symbol in an allele when covering.
     """
    P_dontcare=0.5

    """*
     * The reduction of the prediction error when generating an offspring classifier.
     """
    predictionErrorReduction=0.25
    
    """*
     * The reduction of the fitness when generating an offspring classifier.
     """
    fitnessReduction=0.1

    """*
     * The experience of a classifier required to be a subsumer.
     """
    theta_sub=20
  
    """*
     * The maximal number of steps executed in one trial in a multi-step problem.
     """
    teletransportation=50
  
    """*
     * Specifies if GA subsumption should be executed.
     """
    doGASubsumption=True
    
    """*
     * Specifies if action set subsumption should be executed.
     """
    doActionSetSubsumption=True

    """*
     * The initial prediction value when generating a new classifier (e.g in covering).
     """
    predictionIni=10.0

    """*
     * The initial prediction error value when generating a new classifier (e.g in covering).
     """
    predictionErrorIni=0.0

    """*
     * The initial prediction value when generating a new classifier (e.g in covering).
     """
    fitnessIni=0.01

    """*
     * The don't care symbol (normally '#')
     """
    dontCare='#'
  
    """*
     * The initialization of the pseudo random generator. Must be at lest one and smaller than _M.
     """
    seed=1
    
    """*
     * Constant for the random number generator (modulus of PMMLCG = 2^31 -1).
     """
    _M = 2147483647

    """*
     * Constant for the random number generator (default = 16807).
     """

    _A = 16807
    """*
     * Constant for the random number generator (=_M/_A).
     """

    _Q = _M/_A
    """*
     * Constant for the random number generator (=_M mod _A).
     """
    _R = _M%_A
 
    """*
     * The default constructor.
     """


    """*
     * Sets a random seed in order to randomize the pseudo random generator. 
     """
    def setSeed(self , s):
        self.seed = s

    
    """*
     * Returns a random number in between zero and one.
     """
    def drand(self):
        hi   = self.seed / self._Q
        lo   = self.seed % self._Q
        test = self._A*lo - self._R*hi

    	if test>0:
    	    self.seed = test
    	else:
    	    self.seed = test+self._M

    	return float(self.seed/self._M)
    
    
# we construct this object and the we use it as globals in the other classes
# just by importing this module from XCSConstants import *
XCSConstants = CXCSConstants()
cons = XCSConstants