#-------------------------------------------------------------------------------
# Name:        XCS.py
# Purpose:     Translation of the XCS system to Python
#
# Author:      Jose Antonio Martin H. <jamartin@dia.fi.upm.es>
#
# Created:     16/01/2007
# Copyright:   (c) Jose Antonio Martin H. 2007
# Licence:     <GPL>
#-------------------------------------------------------------------------------
#!/usr/bin/env python
import time
from Environment import *
from XClassifierSet import *
from XCSConstants import *
from PredictionArray import *

    
class XCS:
    """
    This class is the XCS itself. It stores the population and the posed problem.
    The class provides methods for the main learning cycles in XCS distinguishing between
    single-step and multi-step problems as well as exploration vs. exploitation trials.
    Moreover, it handles the performance evaluation.

    @author    Martin V. Butz
    @version   XCSJava 1.0
    Transalted to Python by Jose Antonio Martin H.
    """

    """*
     * Stores the posed problem.
     """
    env = None

    """*
     * Stores the current population of XCS.
     """
    pop = None

    
    """*
     * Stores the specified output File, where the performance will be written.
     """
    outFile = ""
    
    """*
     * Specifies the number of exploration problems/trials to solve in one experiment.
     """
    maxProblems=20000

    """*
     * Specifies the number of investigated experiments.
     """
    nrExps=10

        
    def __init__(self, e, outFileString):
        """*
        * Constructs the XCS system.
        """
        self.env=e

        #specify output file
        self.outFile = outFileString
        
        #initialize XCS
        self.pop=None

        
    def setNumberOfTrials(self,trials):
        """*
        * Resets the maximal number of trials in one experiment. (The default is set to 20000)
        """
    	self.maxProblems=trials

    
        
    def setNumberOfExperiments(self, exps):
        """*
        * Resets the number of experiments. (The default is set to ten.)
        """
        self.nrExps = exps

    
        
    def runXCS(self):
        """*
        * Runs the posed problem with XCS.
        * The function essentially initializes the output File and then runs the experiments.
        *
        * @see #startExperiments
        """

        
        try:
            pW = open(self.outFile,'w')
        except IOError, (errno, strerror):
            print "I/O error(%s): %s" % (errno, strerror)
            raise


        self.startExperiments(pW)
    
        try:
            pW.close()
        except IOError, (errno, strerror):
            print "I/O error(%s): %s" % (errno, strerror)
            raise
  
        
    def startExperiments(self, pW):
        """*
        * This function runs the number of experiments specified. After the initialization of the empty
        * population, either one single- or one multi-step problem is executed.
        *
        * @param pW The print writer refers to the output file for the performance evaluation. The experiments are separated by
        * the text 'Next Experiment'.
        * @see XClassifierSet.XClassifierSet(int)
        * @see Environment.isMultiStepProblem
        * @see doOneSingleStepExperiment()
        * @see doOneMultiStepExperiment()
        """

        for expCounter in range(self.nrExps):
            pW.write("Next Experiment \n")
            print "Experiment Nr."+str(expCounter+1)
            
            #Initialize Population
            self.pop = XClassifierSet(self.env.getNrActions())
           
            if not self.env.isMultiStepProblem():
                self.doOneSingleStepExperiment(pW)
            else:
                self.doOneMultiStepExperiment(pW)

            self.pop=None



    """*************************** Single Step Experiments **************************"""
    
        
    def doOneSingleStepExperiment(self, pW):
        """*
        * Executes one single-step experiment monitoring the performance.
        *
        * @see #doOneSingleStepProblemExplore
        * @see #doOneSingleStepProblemExploit
        * @see #writePerformance
        """

        explore=0
        correct  = [0.0 for i in range(50)]
        sysError = [0.0 for i in range(50)]
        exploreProbC = 0
        while exploreProbC < maxProblems:
            explore = (explore+1)%2
            state = env.resetState()
            
    	    if explore==1:
    	        self.doOneSingleStepProblemExplore(state, exploreProbC)
            else:
                self.doOneSingleStepProblemExploit(state, exploreProbC, correct, sysError)

            if (exploreProbC%50)==0 and explore==0 and exploreProbC>0:
                self.writePerformance(pW, correct, sysError, exploreProbC)

            exploreProbC = exploreProbC + explore



        
    def doOneSingleStepProblemExplore(self,state, counter):
        """*
        * Executes one main learning loop for a single step problem.
        *
        * @see XClassifierSet#XClassifierSet(String,XClassifierSet,int,int)
        * @see PredictionArray#PredictionArray
        * @see PredictionArray#randomActionWinner
        * @see XClassifierSet#XClassifierSet(XClassifierSet,int)
        * @see Environment#executeAction
        * @see XClassifierSet#updateSet
        * @see XClassifierSet#runGA
        * @param state The actual problem instance.
        * @param counter The number of problems observed so far in exploration.
        """

        matchSet        = XClassifierSet(state, self.pop, counter, self.env.getNrActions())
        predictionArray = PredictionArray(matchSet, self.env.getNrActions())
        actionWinner    = predictionArray.randomActionWinner()
        actionSet       = XClassifierSet(matchSet, actionWinner)
        reward          = self.env.executeAction( actionWinner )
        actionSet.updateSet(0.0, reward)
        actionSet.runGA(counter, state, self.env.getNrActions())

  
        
    def doOneSingleStepProblemExploit(self, state, counter, correct, sysError):
        """*
        * Executes one main performance evaluation loop for a single step problem.
        *
        * @see XClassifierSet.XClassifierSet(String,XClassifierSet,int,int)
        * @see PredictionArray.PredictionArray()
        * @see PredictionArray.bestActionWinner()
        * @see Environment.executeAction()
        * @param state The actual problem instance.
        * @param counter The number of problems observed so far in exploration.
        * @param correct The array stores the last fifty correct/wrong exploitation classifications.
        * @param sysError The array stores the last fifty predicted-received reward differences.
        """

        matchSet        = XClassifierSet(state, self.pop, counter, self.env.getNrActions())
        predictionArray = PredictionArray (matchSet, self.env.getNrActions())
        actionWinner    = predictionArray.bestActionWinner()
        reward          = self.env.executeAction( actionWinner )
    
        if self.env.wasCorrect():
            correct[counter%50]=1
        else:
            correct[counter%50]=0
            
        sysError[counter%50] = abs(reward - predictionArray.getBestValue())


    """************************* Multi-step Experiments ******************************"""

        
    def doOneMultiStepExperiment(self,pW):
        """*
        * Executes one multi step experiment and monitors the performance.
        *
        * @see doOneMultiStepProblemExplore()
        * @see doOneMultiStepProblemExploit()
        * @see writePerformance()
        """

        explore            = 0
        exploreStepCounter = 0
        stepsToFood = [0.0 for i in range(50)]
        sysError    = [0.0 for i in range(50)]
        exploreTrialC=0
        
        while exploreTrialC < self.maxProblems:
            explore = (explore+1)%2
            state   = self.env.resetState()
            if explore==1:
                self.exploreStepCounter = self.doOneMultiStepProblemExplore(state, exploreStepCounter)

            else:
                self.doOneMultiStepProblemExploit(state, stepsToFood, sysError, exploreTrialC, exploreStepCounter)

            if (exploreTrialC%50==0) and explore==0 and exploreTrialC>0:
                self.writePerformance(pW, stepsToFood, sysError, exploreTrialC)

            exploreTrialC = exploreTrialC+ explore
           


  
        
    def doOneMultiStepProblemExplore(self, state, stepCounter):
        """*
        * Executes one learning trial in a multi-step problem.
        * A trial normally ends when the food is reached or the teletransportation
        * threshold is reached.
        *
        * @see XCSConstants.teletransportation
        * @see XClassifierSet.XClassifierSet(String,XClassifierSet,int,int)
        * @see PredictionArray.PredictionArray
        * @see PredictionArray.randomActionWinner
        * @see XClassifierSet.XClassifierSet(XClassifierSet,int)
        * @see Environment.executeAction
        * @see XClassifierSet.confirmClassifiersInSet
        * @see XClassifierSet.updateSet
        * @see XClassifierSet.runGA
        * @param state The reseted perception of the problem.
        * @param stepCounter The number of exploration steps executed so far.
        * @return The updated number of exploration setps.
        """

        prevActionSet=None
        prevReward=0.0
        steps=0
        prevState=[]
        
        for steps in range(cons.teletransportation):
            matchSet        = XClassifierSet(state, self.pop, stepCounter+steps, self.env.getNrActions())
            predictionArray = PredictionArray(matchSet, self.env.getNrActions())
            actionWinner    = predictionArray.randomActionWinner()
            actionSet       = XClassifierSet(matchSet, actionWinner)
            reward          = self.env.executeAction(actionWinner)
            

            if prevActionSet!=None:
                prevActionSet.confirmClassifiersInSet()
                prevActionSet.updateSet(predictionArray.getBestValue(), prevReward)
                prevActionSet.runGA(stepCounter+steps, prevState, self.env.getNrActions())


            if self.env.doReset():
                actionSet.confirmClassifiersInSet()
                actionSet.updateSet(0.0, reward)
                actionSet.runGA(stepCounter+steps, state, self.env.getNrActions())
                break

            prevActionSet = actionSet
            prevReward    = reward
            prevState     = state
            state         = self.env.getCurrentState()

        return stepCounter+steps

  
        
    def doOneMultiStepProblemExploit(self, state, stepsToGoal, sysError, trialCounter, stepCounter):
        """*
        * Executes one performance evaluation trial a multi-step problem.
        * Similar to Wilson's exploitation the function updates the parameters of the classifiers
        * but it does not execute the genetic algorithm.
        * A trial normally ends when the food is reached or the teletransportation
        * threshold is reached.
        *
        * @see XCSConstants#teletransportation
        * @see XClassifierSet#XClassifierSet(String,XClassifierSet,int,int)
        * @see PredictionArray#PredictionArray
        * @see PredictionArray#bestActionWinner
        * @see XClassifierSet#XClassifierSet(XClassifierSet,int)
        * @see Environment#executeAction
        * @see XClassifierSet#confirmClassifiersInSet
        * @see XClassifierSet#updateSet
        * @param state The reseted perception of the problem.
        * @param stepsToGoal The last fifty numbers of steps to the goal during exploitation.
        * @param sysError The averaged prediction errors of the last fifty trials.
        * @param trialCounter The number of exploration trials executed so far.
        * @param stepCounter The number of exploration steps executed so far.
        """

        prevActionSet  = None
        prevReward     = 0.0
        prevPrediction = 0.0
        nsteps         = 1.0
        sysError[trialCounter%50] = 0.0

        for steps in range(cons.teletransportation):
            matchSet = XClassifierSet(state, self.pop, stepCounter, self.env.getNrActions())
            predictionArray = PredictionArray(matchSet, self.env.getNrActions())
            actionWinner = predictionArray.bestActionWinner()
            actionSet = XClassifierSet(matchSet, actionWinner)
            reward = self.env.executeAction( actionWinner )
      
            if prevActionSet!=None:
                prevActionSet.confirmClassifiersInSet();
                prevActionSet.updateSet(predictionArray.getBestValue(), prevReward);
                sysError[trialCounter%50] = sysError[trialCounter%50] + abs(cons.gamma * predictionArray.getValue(actionWinner) + prevReward - prevPrediction) / float(self.env.getMaxPayoff())


            if self.env.doReset():
                actionSet.confirmClassifiersInSet()
                actionSet.updateSet(0.0, reward)
                sysError[trialCounter%50] = sysError[trialCounter%50] + abs(reward - predictionArray.getValue(actionWinner)) / float(self.env.getMaxPayoff())
                break

            prevActionSet   = actionSet
            prevPrediction  = predictionArray.getValue(actionWinner)
            prevReward      = reward
            state           = self.env.getCurrentState()
            nsteps = steps + 1
        sysError[trialCounter%50] = sysError[trialCounter%50] / float(nsteps)
        stepsToGoal[trialCounter%50]=nsteps
        

  
    """##########---- Output ----##########"""

        
    def writePerformance(self, pW, performance, sysError, exploreProbC):
        """*
        * Writes the performance of the XCS to the specified file.
        * The function writes time performance systemError actualPopulationSizeInMacroClassifiers.
        * Performance and system error are averaged over the last fifty trials.
        *
        * @param pW The reference where to write the performance.
        * @param performance The performance in the last fifty exploration trials.
        * @param sysError The system error in the last fifty exploration trials.
        * @param exploreProbC The number of exploration trials executed so far.
        """

        perf  = sum(performance)/50.0
        serr  = sum(sysError)/50.0
        
        pW.write("" + str(exploreProbC) + " " + str(perf) + " " + str(serr) + " " + str(self.pop.getSize()) + "\n")
        print "Probe: " + str(exploreProbC) + "\t Performance: " + str(perf) + "\t Error: " + str(serr) + "\t PopSize: " + str(self.pop.getSize())


